%{
int parenth_list = 0;
int parenth_exp  = 0;
int absorb       = 1;
char *code_file_name = "";
int line_number = 0;

static int parse_string = 0;
static int in_comment   = 0;

static void         comment            (void);
static char        *exhaust            (void);
static void         directive          (void);
static char        *parse_parenth_exp  (void);
static char        *c_block            (void);
static item_list_t  parse_parenth_list (void);
%}
%x SPECIAL ABSORB

alpha   [a-zA-Z]
digit   [0-9]
special [\_]
ident   (({alpha}|{special})({alpha}|{digit}|{special})*)

%%
                if (absorb) BEGIN(ABSORB);

<SPECIAL>{
"\n"            {line_number++; ECHO;}
"/*"            {comment ();}
"//"            {exhaust ();}
"#"             {directive ();}

[\ \t\f\n]      {}
module          {return MODULE;}
native          {return NATIVE;}
run             {yylval.intval = line_number; return RUN;}
local           {return LOCAL;}
finalize        {return FINALIZE;}

cooperate       {yylval.intval = line_number; return COOPERATE;}
halt            {yylval.intval = line_number; return HALT;}
while           {yylval.intval = line_number; return WHILE_TOK;}
repeat          {yylval.intval = line_number; return REPEAT;}
do              {return DO;}
end             {return END;}
if              {yylval.intval = line_number; return IF;}
then            {yylval.intval = line_number; return THEN;}
else            {yylval.intval = line_number; return ELSE;}
 
await           {yylval.intval = line_number; return AWAIT;}
get_value       {yylval.intval = line_number; return GET_VALUE;}

join            {yylval.intval = line_number; return JOIN;}

return          {yylval.intval = line_number; return RETURN;}
goto            {yyerror("GOTO ERROR");}
break           {yyerror("BREAK ERROR");}

unlink          {yylval.intval = line_number; return UNLINK;}
link            {yylval.intval = line_number; return LINK;}

"{"             {yylval.str = c_block (); return C_STAT;}

"("             {
                    if (parenth_exp){yylval.str = parse_parenth_exp (); return PARENTH_EXP;}
                    if (parenth_list){yylval.item_list = parse_parenth_list (); return PARENTH_LIST;}
                    return yytext[0];
                }

{ident}         {yylval.str = mysprintf("%s",yytext); return IDENTIFIER;}

.               {return yytext[0];}
}
<ABSORB>{
"\n"            {line_number++; ECHO;}
"module "       {if (!in_comment && !parse_string){absorb=0;BEGIN(SPECIAL);}else ECHO;}

"/*"            {comment();}
"//"            {exhaust ();}
"#"             {directive ();}

"\""            {parse_string = !parse_string; ECHO;}
"\\\""          {ECHO;}

.               {ECHO;}
}

%%
int yywrap() {return(1);}

int yyerror (char *msg)
{
   fprintf(stderr,"error line %d: %s (last token: %s)\n",line_number,msg,yytext);
   exit(1);
   return 0;
}

static void comment(void)
{
    int init = line_number;
    char c;
    in_comment = 1;
    while(1){
       while ( (c = input()) != '*' && c != EOF ) /* eat */; 
        if ( c == '*' ){
            while ( (c = input()) == '*' );
            if ( c == '/' ) break; /* found the end */
        }
        if ( c == EOF ){
            yyerror (mysprintf ("EOF in comment starting at line %d",init));
            break;
        }
   }
   in_comment = 0;
}

static char *exhaust (void)
{
    char c, *res = "";
    while ((c = input()) != '\n' && c != EOF) res = mysprintf ("%s%c",res,c);
    if (c == EOF) yyerror ("unexpected EOF");
    return res;
}

static void directive (void)
{
   source (exhaust ());
}

static char *c_block (void)
{
    int n = 1, init = line_number;
    char c, *res = "";
    while(1){
       while ( (c = input()) != '}' && c != '{' && c != EOF ) {
         res = mysprintf ("%s%c",res,c);
       }
       if (c == '{') {
         n++; 
       } else if (c == '}') {
          if (n == 1) break; else n--;
       } else if (c == EOF) {
          yyerror (mysprintf ("EOF in atomic statement starting at line %d",
                   init));
       }
       res = mysprintf ("%s%c",res,c);
   }
   return mysprintf ("%s",res);
}

static char *parse_parenth_exp (void)
{
    int n = 1, init = line_number;
    char c, *res = "";
    while(1){
       while ( (c = input()) != ')' && c != '(' && c != EOF ) {
         res = mysprintf ("%s%c",res,c);
       }
       if (c == '(') {
         n++; 
       } else if (c == ')') {
          if (n == 1) break; else n--;
       } else if (c == EOF) {
          yyerror (mysprintf ("EOF in parenthesized expression at line %d",init));
       }
       res = mysprintf ("%s%c",res,c);
   }
   return mysprintf ("%s",res);
}

static item_list_t parse_parenth_list (void)
{
    int n = 1, init = line_number;
    char c, *item = "";
    item_list_t res = _make_item_list ();
    while(1){
       while ( (c = input()) != ')' && c != '(' && c != ',' && c != EOF ) {
         item = mysprintf ("%s%c",item,c);
       }
       if (c == '(') {
         n++; 
       } else if (c == ',' && n == 1) {
         _append_item_list (res,item);
         item = "";
         continue;
       } else if (c == ')') {
          if (n == 1) {
             _append_item_list (res,item);
             break;
          }
          n--;
       } else if (c == EOF) {
          yyerror (mysprintf ("EOF in parenthesized list at line %d",init));
       }
       item = mysprintf ("%s%c",item,c);
   }
   return res;
}
