#include "status.h"
#include "loft_sem.h"

extern void add_to_broadcast (scheduler_t,broadcast_t);

/***************************************************/
event_t event_create_in (scheduler_t sched)
{
   event_t res    = mymalloc (sizeof (struct _event_t));
   res->values    = _make_item_list ();
   res->scheduler = sched;
#ifdef OPTIM   
   res->waiting   = _make_item_list ();
#endif
   return res;
}

event_t event_create (void)
{
   return event_create_in (implicit_scheduler ());
}

/***************************************************/
void reset_event_values (event_t event)
{
   _reset_item_list (event->values);
}

void add_event_value (event_t event,void *val)
{
   _append_item_list (event->values,val);
}

int event_values_length (event_t event)
{
   return _length_of_item_list (event->values);
}

void *nth_of_event_values (event_t event,int k)
{
   return _nth_of_item_list (event->values,k);
}

/*********************************************/
#ifdef OPTIM
void register_in_event (event_t event)
{
   _append_item_list (event->waiting,self ());
}

static void awake (event_t event)
{
   item_cell_t cell = _first_cell (event->waiting);
   while (cell) {
      thread_t th = _content_of_cell (cell);
      if (!th->suspended && th->state == _WAIT) th->state = _CONT;
      cell = _next_cell (cell);
   }
   _reset_item_list (event->waiting);
}
#endif

/***************************************************/
int is_present (event_t event)
{
   item_list_t list = event->scheduler->events;
   item_cell_t cell = _first_cell (list);
   while (cell != NULL) {
      if (_content_of_cell (cell) == event) return 1;
      cell = _next_cell (cell);
   }
   return 0;
}

void broadcast (event_t event)
{
   if (!is_present (event)) {
      scheduler_t sched = event->scheduler;
      reset_event_values (event);
      _append_item_list (sched->events,event);
      sched->move = 1;
#ifdef OPTIM
      awake (event);
#endif
   }
}

void broadcast_value (event_t event,void *val)
{
   broadcast (event);
   add_event_value (event,val);
   event->scheduler->move = 1;
}

/***************************************************/
void generate (event_t event)
{
   scheduler_t sched = event->scheduler;
   if (current_scheduler () != sched) {
      add_to_broadcast (sched,make_broadcast (event,1,NULL));
   } else {
      broadcast (event);
   }
}

void generate_value (event_t event,void *val)
{
   scheduler_t sched = event->scheduler;
   if (current_scheduler () != sched) {
      add_to_broadcast (sched,make_broadcast (event,0,val));
   } else {
      broadcast_value (event,val);
   }
}
