#include "replace_sem.h"

/********************************************/
#define SET_TYPE(x,v) x->parent.type = v

/********************************************/
void cooperate_reset (instruction_t inst)
{
   cooperate_t coop = (cooperate_t) inst;
   coop->state = 0;
}

instruction_t make_cooperate (void)
{
   cooperate_t res = mymalloc (sizeof (struct _cooperate_t));
   SET_TYPE(res,_COOPERATE);
   res->state = 0;
   return (instruction_t)res;
}

/********************************************/
static struct _instruction_t nothing_ = {_NOTHING};

instruction_t make_nothing (void)
{
   return &nothing_;
}

/********************************************/
static struct _instruction_t halt_ = {_HALT};

instruction_t make_halt ()
{
   return &halt_;
}

/********************************************/
static struct _instruction_t return_ = {_RETURN};

instruction_t make_return ()
{
   return &return_;
}

/********************************************/
void sequence_reset (instruction_t inst)
{
   sequence_t s = (sequence_t)inst;
   s->state = 0;
   reset_instruction (s->left);
   reset_instruction (s->right);   
}
instruction_t make_sequence (instruction_t left,instruction_t right)
{
   sequence_t res = mymalloc (sizeof (struct _sequence_t));
   SET_TYPE(res,_SEQ);
   res->state = 0;
   res->left = left;
   res->right = right;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_stop (expression_t exp)
{
   stop_t res = mymalloc (sizeof (struct _stop_t));
   SET_TYPE(res,_STOP);
   res->exp = exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_suspend (expression_t exp)
{
   suspend_t res = mymalloc (sizeof (struct _suspend_t));
   SET_TYPE(res,_SUSPEND);
   res->exp = exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_resume (expression_t exp)
{
   resume_t res = mymalloc (sizeof (struct _resume_t));
   SET_TYPE(res,_RESUME);
   res->exp = exp;
   return (instruction_t)res;
}

/********************************************/
void if_reset (instruction_t inst)
{
   if_t i = (if_t)inst;
   i->state = 0;
   reset_instruction (i->left);
   reset_instruction (i->right);
}

instruction_t make_if (expression_t exp,instruction_t left,instruction_t right)
{
   if_t res = mymalloc (sizeof (struct _if_t));
   SET_TYPE(res,_IF);
   res->exp = exp;
   res->state = 0;
   res->left = left;
   res->right = right;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_inst_from_exp (expression_t exp)
{
   inst_from_exp_t res = mymalloc (sizeof (struct _inst_from_exp_t));
   SET_TYPE(res,_EXP_INST);
   res->exp = exp;
   return (instruction_t)res;
}

/********************************************/
void repeat_reset (instruction_t inst)
{
   repeat_t rep = (repeat_t)inst;
   rep->state = 0;
   reset_instruction (rep->body);
}

instruction_t make_repeat (expression_t exp,instruction_t body)
{
   repeat_t res = mymalloc (sizeof (struct _repeat_t));
   SET_TYPE(res,_REPEAT);
   res->exp = exp;
   res->body = body;
   res->state = 0;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_generate (expression_t exp)
{
   generate_t res = mymalloc (sizeof (struct _generate_t));
   SET_TYPE(res,_GENERATE);
   res->exp = exp;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_generate_valued (expression_t evt_exp,expression_t val_exp)
{
   generate_valued_t res = mymalloc (sizeof (struct _generate_valued_t));
   SET_TYPE(res,_GENERATE_VALUED);
   res->evt_exp = evt_exp;
   res->val_exp = val_exp;
   return (instruction_t)res;
}

/********************************************/
void await_reset (instruction_t inst)
{
   await_t a = (await_t)inst;
   a->state = 0;
}

instruction_t make_await (expression_t exp)
{
   await_t res = mymalloc (sizeof (struct _await_t));
   SET_TYPE(res,_AWAIT);
   res->exp = exp;
   res->state = 0;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_limited_await (expression_t evt_exp,expression_t limit_exp)
{
   limited_await_t res = mymalloc (sizeof (struct _limited_await_t));
   SET_TYPE(res,_LIMITED_AWAIT);
   res->evt_exp = evt_exp;
   res->limit_exp = limit_exp;
   res->state = 0;
   return (instruction_t)res;
}

/********************************************/
instruction_t make_atom (void *(*selector)(),int n)
{
   atom_t res = mymalloc (sizeof (struct _atom_t));
   SET_TYPE(res,_ATOM);
   res->selector = selector;
   res->n = n;
   return (instruction_t)res;
}

/********************************************/
void join_reset (instruction_t inst)
{
   join_t j = (join_t)inst;
   j->state = 0;
}

instruction_t make_join (expression_t exp)
{
   join_t res = mymalloc (sizeof (struct _join_t));
   SET_TYPE(res,_JOIN);
   res->exp = exp;
   res->state = 0;
   return (instruction_t)res;
}

/********************************************/
void limited_join_reset (instruction_t inst)
{
   limited_join_t j = (limited_join_t)inst;
   j->state = 0;
}

instruction_t make_limited_join (expression_t th_exp,expression_t limit_exp)
{
   limited_join_t res = mymalloc (sizeof (struct _limited_join_t));
   SET_TYPE(res,_LIMITED_JOIN);
   res->th_exp = th_exp;
   res->limit_exp = limit_exp;
   res->state = 0;
   return (instruction_t)res;
}

/********************************************/
void run_reset (instruction_t inst)
{
   run_t r = (run_t)inst;
   r->state = 0;
}

instruction_t make_run (expression_t exp)
{
   run_t res = mymalloc (sizeof (struct _run_t));
   SET_TYPE(res,_RUN);
   res->exp = exp;
   res->state = 0;
   return (instruction_t)res;
}

/********************************************/
void while_reset (instruction_t inst)
{
   while_t w = (while_t)inst;
   w->state = 0;
   reset_instruction (w->body);
}

instruction_t make_while (expression_t exp,instruction_t body)
{
   while_t res = mymalloc (sizeof (struct _while_t));
   SET_TYPE(res,_WHILE);
   res->exp = exp;
   res->body = body;
   res->state = 0;
   return (instruction_t)res;
}

/********************************************/
void get_value_reset (instruction_t inst)
{
   get_value_t g = (get_value_t)inst;
   g->state = 0;
}

instruction_t make_get_value (expression_t evt_exp,expression_t num_exp,void **var)
{
   get_value_t res = mymalloc (sizeof (struct _get_value_t));
   SET_TYPE(res,_GET_VALUE);
   res->evt_exp = evt_exp;
   res->num_exp = num_exp;
   res->var = var;
   res->state = 0;
   return (instruction_t)res;
}

instruction_t make_automaton (aut_fun_t fun)
{
   automaton_t res = mymalloc (sizeof (struct _automaton_t));
   SET_TYPE(res,_AUTOMATON);
   res->fun = fun;
   return (instruction_t)res;
}

/********************************************/
void no_reset (instruction_t inst) {}

reset_t dispach_reset[] = {
      no_reset              ,
      cooperate_reset       ,
      no_reset              ,
      no_reset              ,
      sequence_reset        ,
      no_reset              ,
      no_reset              ,
      no_reset              ,
      if_reset              ,
      no_reset              ,
      repeat_reset          ,
      no_reset              ,
      no_reset              ,
      await_reset           ,
      await_reset           ,
      no_reset              ,
      join_reset            ,
      limited_join_reset    ,
      run_reset             ,
      while_reset           ,
      get_value_reset       ,
      no_reset              ,
};

void reset_instruction (instruction_t inst)
{
   return dispach_reset [(int)inst->type] (inst);
}

