#define _NOTHING                  0
#define _COOPERATE                1
#define _HALT                     2
#define _RETURN                   3
#define _SEQ                      4
#define _STOP                     5
#define _SUSPEND                  6
#define _RESUME                   7
#define _IF                       8
#define _EXP_INST                 9
#define _REPEAT                  10
#define _GENERATE                11
#define _GENERATE_VALUED         12
#define _AWAIT                   13
#define _LIMITED_AWAIT           14
#define _ATOM                    15
#define _JOIN                    16
#define _LIMITED_JOIN            17
#define _RUN                     18
#define _WHILE                   19
#define _GET_VALUE               20

#define _REPEAT_STAR             21
#define _WHILE_STAR              22
#define _LIMITED_AWAIT_STAR      23
#define _AWAIT_STAR              24
#define _GET_VALUE_STAR          25



#define PACK __attribute__ ((packed))
//#define PACK

/*****************************/
typedef struct _instruction_t
{
   char type;
}
*instruction_t;

/*****************************/
typedef struct _sequence_t
{
   struct _instruction_t parent;
   instruction_t left;
   instruction_t right;
} PACK
*sequence_t;

/*****************************/
typedef struct _stop_t
{
   struct _instruction_t parent;
   expression_t exp;
} PACK
*stop_t;

/*****************************/
typedef struct _suspend_t
{
   struct _instruction_t parent;
   expression_t exp;
} PACK
*suspend_t;

/*****************************/
typedef struct _resume_t
{
   struct _instruction_t parent;
   expression_t exp;
} PACK
*resume_t;

/*****************************/
typedef struct _if_t
{
   struct _instruction_t parent;
   expression_t exp;
   instruction_t left;
   instruction_t right;   
} PACK
*if_t;

/*****************************/
typedef struct _inst_from_exp_t
{
   struct _instruction_t parent;
   expression_t exp;
} PACK
*inst_from_exp_t;

/*****************************/
typedef struct _repeat_t
{
   struct _instruction_t parent;
   expression_t exp;
   instruction_t body;
   int counter;
} PACK
*repeat_t;

/*****************************/
typedef struct _repeat_star_t
{
   struct _instruction_t parent;
   instruction_t init;
   instruction_t body;
   int counter;
} PACK
*repeat_star_t;

/*****************************/
typedef struct _generate_t
{
   struct _instruction_t parent;
   expression_t exp;
} PACK
*generate_t;

/*****************************/
typedef struct _generate_valued_t
{
   struct _instruction_t parent;
   expression_t evt_exp;
   expression_t val_exp;   
} PACK
*generate_valued_t;

/*****************************/
typedef struct _await_t
{
   struct _instruction_t parent;
   expression_t exp;
   event_t event;
} PACK
*await_t;

/*****************************/
typedef struct _await_star_t
{
   struct _instruction_t parent;
   event_t event;
} PACK
*await_star_t;

/*****************************/
typedef struct _limited_await_t
{
   struct _instruction_t parent;
   expression_t evt_exp;
   expression_t limit_exp;   
   event_t event;
   int limit;
} PACK
*limited_await_t;

typedef struct _limited_await_star_t
{
   struct _instruction_t parent;
   event_t event;
   int limit;
} PACK
*limited_await_star_t;

/*****************************/
typedef struct _atom_t
{
   struct _instruction_t parent;
   void *(*selector)();
   int n;
} PACK
*atom_t;

/*****************************/
typedef struct _join_t
{
   struct _instruction_t parent;
   expression_t exp;
   thread_t thread;
} PACK
*join_t;

/*****************************/
typedef struct _limited_join_t
{
   struct _instruction_t parent;
   expression_t th_exp;
   thread_t thread;
   expression_t limit_exp;
   int limit;
} PACK
*limited_join_t;

/*****************************/
typedef struct _run_t
{
   struct _instruction_t parent;
   expression_t exp;
   thread_t thread;
} PACK
*run_t;

/*****************************/
typedef struct _while_t
{
   struct _instruction_t parent;
   expression_t exp;
   instruction_t body;
} PACK
*while_t;

/*****************************/
typedef struct _while_star_t
{
   struct _instruction_t parent;
   expression_t exp;
   instruction_t init;
   instruction_t body;
} PACK
*while_star_t;

/*****************************/
typedef struct _get_value_t
{
   struct _instruction_t parent;
   expression_t evt_exp;
   event_t event;
   expression_t num_exp;
   int num;
   void **var;
} PACK
 *get_value_t ;

/*****************************/
typedef struct _get_value_star_t
{
   struct _instruction_t parent;
   event_t event;
   int num;
   void **var;
} PACK
 *get_value_star_t ;

/*****************************/
#include "inst_make.h"

instruction_t make_repeat_star        (int n,instruction_t init,instruction_t body);
instruction_t make_await_star         (event_t);
instruction_t make_limited_await_star (event_t,int n);
instruction_t make_while_star         (expression_t,instruction_t init,instruction_t body);
instruction_t make_get_value_star     (event_t,int,void **var);
